--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, diag_lib, mm_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY tb_correlator IS
END tb_correlator;

ARCHITECTURE tb OF tb_correlator IS

  CONSTANT c_nof_inputs         : NATURAL := 24;
  CONSTANT c_nof_mults          : NATURAL := (c_nof_inputs*(c_nof_inputs+1))/2;

  CONSTANT c_clk_period         : TIME := 10 ns;
  CONSTANT c_data_w             : NATURAL := 32;

  -- Block generator
  CONSTANT c_bg_block_size              : NATURAL := 1024;
  CONSTANT c_bg_gapsize                 : NATURAL := 0;
  CONSTANT c_bg_blocks_per_sync         : NATURAL := 10;
  CONSTANT c_bg_ctrl                    : t_diag_block_gen := ('1',                      -- enable             
                                                               '0',                      -- enable_sync        
                                                              TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                              TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                              TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                              TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                              TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                              TO_UVEC(                   0, c_diag_bg_bsn_init_w));


  SIGNAL tb_end                 : STD_LOGIC := '0';
  SIGNAL clk                    : STD_LOGIC := '1';
  SIGNAL rst                    : STD_LOGIC;

  SIGNAL correlator_snk_in_arr  : t_dp_sosi_arr(c_nof_inputs-1 DOWNTO 0);
  SIGNAL correlator_src_out_arr : t_dp_sosi_arr(c_nof_inputs*(c_nof_inputs+1)/2-1 DOWNTO 0);

  SIGNAL ram_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL ram_diag_data_buf_miso : t_mem_miso;
  SIGNAL reg_diag_data_buf_mosi : t_mem_mosi;
  SIGNAL reg_diag_data_buf_miso : t_mem_miso; 

  ----------------------------------------------------------------------------
  -- mm_file component
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;
    g_mm_clk_period     : TIME := c_clk_period;
    g_update_on_change  : BOOLEAN := FALSE;
    g_mm_rd_latency     : NATURAL := 1
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;

BEGIN

  -----------------------------------------------------------------------------
  -- Clocks and reset
  -----------------------------------------------------------------------------
  clk <= NOT clk OR tb_end AFTER c_clk_period/2;
  rst <= '1', '0' AFTER c_clk_period*7;

  -----------------------------------------------------------------------------
  -- Block generators
  -- . Each stream in block_gen_src_out_arr contains complex subband samples
  --   from one 'antenna'.
  -- . These complex subband samples are generated and converted to a HEX RAM
  --   initialization file using Python, see tb/python/gen_subband_hex_files.py
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP (
    g_nof_output_streams => c_nof_inputs,
    g_buf_dat_w          => c_data_w,
    g_buf_addr_w         => ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
    g_file_name_prefix   => "../../../libraries/dsp/correlator/src/hex/complex_subbands_" & NATURAL'IMAGE(c_data_w/2),
    g_diag_block_gen_rst => c_bg_ctrl
  )
  PORT MAP (
    mm_rst           => rst,
    mm_clk           => clk,

    dp_rst           => rst,
    dp_clk           => clk,

    out_sosi_arr     => correlator_snk_in_arr
  );

  -----------------------------------------------------------------------------
  -- Device under test: correlator
  -----------------------------------------------------------------------------
  u_correlator : ENTITY work.correlator
  GENERIC MAP (
    g_nof_inputs => c_nof_inputs,
    g_nof_mults  => c_nof_mults
  )
  PORT MAP (
    clk         => clk,
    rst         => rst,

    snk_in_arr  => correlator_snk_in_arr,
    src_out_arr => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Data buffers to be read out by Python
  -----------------------------------------------------------------------------
  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams  => c_nof_mults,
    g_data_w       => c_data_w,
    g_data_type    => e_complex,
    g_buf_nof_data => c_bg_block_size,
    g_buf_use_sync => FALSE 
  )
  PORT MAP (
    mm_rst            => rst,
    mm_clk            => clk,
    dp_rst            => rst,
    dp_clk            => clk,

    ram_data_buf_mosi => ram_diag_data_buf_mosi,
    ram_data_buf_miso => ram_diag_data_buf_miso,
    reg_data_buf_mosi => reg_diag_data_buf_mosi,
    reg_data_buf_miso => reg_diag_data_buf_miso,

    in_sosi_arr       => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- MM file I/O with Python
  -----------------------------------------------------------------------------
  u_mm_file_ram_diag_data_buffer : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_DATA_BUFFER")
                                              PORT MAP(rst, clk, ram_diag_data_buf_mosi, ram_diag_data_buf_miso);

  u_mm_file_reg_diag_data_buffer : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_DATA_BUFFER")
                                              PORT MAP(rst, clk, reg_diag_data_buf_mosi, reg_diag_data_buf_miso);

  ----------------------------------------------------------------------------
  -- Procedure that polls a sim control file that can be used to e.g. get
  -- the simulation time in ns
  ----------------------------------------------------------------------------
  mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");

END tb;
