-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, diag_lib, tr_xaui_lib, technology_lib, tech_mac_10g_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE technology_lib.technology_pkg.ALL;

ENTITY tr_10GbE IS
  GENERIC (
    g_technology        : NATURAL := c_tech_select_default;
    g_sim               : BOOLEAN;
    g_sim_level         : NATURAL := 0; -- 0 = use IP; 1 = use fast serdes model
    g_nof_macs          : NATURAL;
    g_use_mdio          : BOOLEAN;
    g_mdio_epcs_dis     : BOOLEAN := FALSE; -- TRUE disables EPCS on init; e.g. to target a 10GbE card in PC that does not support it 
    g_pkt_len           : NATURAL := 100;
    g_lpbk_sosi         : BOOLEAN := FALSE; --  i/o pins <-> ][ <-> tr_xaui <-> ][ <-> tr_10GbE <-> ][ <-> packetizing <-> user
    g_lpbk_xgmii        : BOOLEAN := FALSE; --               ^^                 ^^                  ^^ 
    g_lpbk_xaui         : BOOLEAN := FALSE; --    g_lpbk_xaui|      g_lpbk_xgmii|        g_lpbk_sosi|
    g_use_hdr_ram       : BOOLEAN := FALSE;
    g_hdr_ram_init_file : STRING := "UNUSED";
    g_hdr_release_at_init : STD_LOGIC := '0'; -- '1' = Release header onto datapath at init
    g_word_alignment_padding : BOOLEAN := FALSE 
  );
  PORT (
    mm_rst              : IN  STD_LOGIC;
    mm_clk              : IN  STD_LOGIC;

    tr_clk              : IN  STD_LOGIC;

    cal_rec_clk         : IN  STD_LOGIC;

    dp_rst              : IN  STD_LOGIC := '0';
    dp_clk              : IN  STD_LOGIC := '0';

    snk_out_arr         : OUT t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);
    snk_in_arr          : IN  t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0) := (OTHERS=>c_dp_sosi_rst);

    src_in_arr          : IN  t_dp_siso_arr(g_nof_macs-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr         : OUT t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0);

    -- MM registers
    reg_hdr_insert_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_hdr_insert_mosi : IN  t_mem_mosi := c_mem_mosi_rst;

    ram_hdr_remove_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_hdr_remove_miso : OUT t_mem_miso := c_mem_miso_rst;

    reg_mac_mosi        : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_mac_miso        : OUT t_mem_miso := c_mem_miso_rst;

    --Serial I/O
    xaui_rx_in_arr      : IN  t_xaui_arr(g_nof_macs-1 DOWNTO 0);
    xaui_tx_out_arr     : OUT t_xaui_arr(g_nof_macs-1 DOWNTO 0);   

    -- Serial IO for Arria
    tx_serial_data      : out std_logic_vector(g_nof_macs-1 downto 0);
    rx_serial_data      : in  std_logic_vector(g_nof_macs-1 downto 0);
  
    -- MDIO External clock and serial data i/o
    mdio_rst            : OUT STD_LOGIC;
    mdio_mdc_arr        : OUT STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);       
    mdio_mdat_in_arr    : IN  STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0) := (OTHERS=>'0');
    mdio_mdat_oen_arr   : OUT STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);

    mdio_mosi_arr       : IN  t_mem_mosi_arr(g_nof_macs-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
    mdio_miso_arr       : OUT t_mem_miso_arr(g_nof_macs-1 DOWNTO 0);

    -- XAUI PHY IP MM control/status
    xaui_mosi           : IN  t_mem_mosi := c_mem_mosi_rst;
    xaui_miso           : OUT t_mem_miso
  );
END tr_10GbE;


ARCHITECTURE str OF tr_10GbE IS

  CONSTANT c_nof_header_words       : NATURAL := 8; --8 64b words = 16 32b words

  CONSTANT c_mac_mm_addr_w          : NATURAL := 13;

  CONSTANT c_hdr_insert_reg_addr_w  : NATURAL := 1; -- Only 1 register used. A width of 1 still yields 2 addresses/instance though.
  CONSTANT c_hdr_insert_ram_addr_w  : NATURAL := ceil_log2( c_nof_header_words * (c_xgmii_data_w/c_word_w) );
  CONSTANT c_hdr_remove_ram_addr_w  : NATURAL := ceil_log2( c_nof_header_words * (c_xgmii_data_w/c_word_w) );

  CONSTANT c_fifo_margin            : NATURAL := g_pkt_len;

  CONSTANT c_word_alignment_padding_nof_bytes : NATURAL := 6; -- 6 Bytes (=48 bits) of padding aligns the 336 bit eth/ip/udp header to a 64b word boundary: 336+48=384 (multiple of 64).

  CONSTANT c_remove_crc             : BOOLEAN := sel_a_b(g_lpbk_sosi, FALSE, TRUE); --The MAC adds a CRC
  CONSTANT c_nof_crc_words          : NATURAL := (c_network_eth_crc_len * c_byte_w) / c_word_w;

  -- Tail: One packet contianing only one word of which the last 4 (LS) bytes are empty
  CONSTANT c_empty_crc_sosi         : t_dp_sosi := ('0', (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), '1', '1', '1', (OTHERS=>'1'), (OTHERS=>'0'), (OTHERS=>'0'));

  SIGNAL tx_clk_arr                     : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);
  SIGNAL tx_rst_arr                     : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);

  SIGNAL rx_clk_arr                     : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_rst_arr                     : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);

  SIGNAL tx_rst_n_arr                   : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_rst_n_arr                   : STD_LOGIC_VECTOR(g_nof_macs-1 DOWNTO 0);

  SIGNAL mm_rst_n                       : STD_LOGIC;

  SIGNAL reg_hdr_insert_mosi_arr        : t_mem_mosi_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL ram_hdr_insert_mosi_arr        : t_mem_mosi_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL ram_hdr_remove_mosi_arr        : t_mem_mosi_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL ram_hdr_remove_miso_arr        : t_mem_miso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL reg_mac_mosi_arr               : t_mem_mosi_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL reg_mac_miso_arr               : t_mem_miso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_fifo_dc_tx_src_out_arr      : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_fifo_dc_tx_src_in_arr       : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_hdr_insert_src_out_arr      : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL dp_hdr_insert_src_in_arr       : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_pad_remove_src_out_arr      : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL dp_pad_remove_src_in_arr       : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_concat_snk_in_2arr          : t_dp_sosi_2arr_2(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_concat_snk_out_2arr         : t_dp_siso_2arr_2(g_nof_macs-1 DOWNTO 0); 

  SIGNAL dp_concat_src_out_arr          : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_concat_src_in_arr           : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0); 

  SIGNAL dp_fifo_fill_snk_in_arr        : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_fifo_fill_snk_out_arr       : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0); 

  SIGNAL dp_fifo_fill_src_out_arr       : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_fifo_fill_src_in_arr        : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_latency_adapter_snk_in_arr  : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_latency_adapter_snk_out_arr : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL rx_sosi_arr                    : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL rx_siso_arr                    : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_frame_remove_snk_in_arr     : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_frame_remove_snk_out_arr    : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_pad_insert_snk_in_arr       : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_pad_insert_snk_out_arr      : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL mac_10g_snk_in_arr             : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL mac_10g_snk_out_arr            : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL mac_10g_src_out_arr            : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL mac_10g_src_in_arr             : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_fifo_dc_rx_snk_in_arr       : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_fifo_dc_rx_snk_out_arr      : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL dp_fifo_dc_rx_src_out_arr      : t_dp_sosi_arr(g_nof_macs-1 DOWNTO 0); 
  SIGNAL dp_fifo_dc_rx_src_in_arr       : t_dp_siso_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL mac_xgmii_tx_dc_arr            : t_xgmii_dc_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL mac_xgmii_rx_dc_arr            : t_xgmii_dc_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL xaui_xgmii_tx_dc_arr           : t_xgmii_dc_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL xaui_xgmii_rx_dc_arr           : t_xgmii_dc_arr(g_nof_macs-1 DOWNTO 0);

  SIGNAL xaui_rx_arr                    : t_xaui_arr(g_nof_macs-1 DOWNTO 0);
  SIGNAL xaui_tx_arr                    : t_xaui_arr(g_nof_macs-1 DOWNTO 0);  

  -- signals for arria 10
  SIGNAL tx_serial_data_internal        : std_logic_vector(g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_serial_data_internal        : std_logic_vector(g_nof_macs-1 DOWNTO 0);
  SIGNAL tx_parallel_data               : std_logic_vector(64*g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_parallel_data               : std_logic_vector(64*g_nof_macs-1 DOWNTO 0);
  SIGNAL tx_control                     : std_logic_vector(8*g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_control                     : std_logic_vector(8*g_nof_macs-1 DOWNTO 0);
  SIGNAL tx_data_valid                  : std_logic_vector(g_nof_macs-1 DOWNTO 0);
  SIGNAL rx_data_valid                  : std_logic_vector(g_nof_macs-1 DOWNTO 0);


  ---------------------------------------------------------------------------------------
  -- Component declarations to prevent synthesis of unused technologies
  ---------------------------------------------------------------------------------------

  component ip_arria10_mac_10g is
  port (
    csr_read                        : in  std_logic                     := 'X';             -- read
    csr_write                       : in  std_logic                     := 'X';             -- write
    csr_writedata                   : in  std_logic_vector(31 downto 0) := (others => 'X'); -- writedata
    csr_readdata                    : out std_logic_vector(31 downto 0);                    -- readdata
    csr_waitrequest                 : out std_logic;                                        -- waitrequest
    csr_address                     : in  std_logic_vector(12 downto 0)  := (others => 'X'); -- address
    tx_312_5_clk                    : in  std_logic                     := 'X';             -- clk
    tx_156_25_clk                   : in  std_logic                     := 'X';             -- clk
    rx_312_5_clk                    : in  std_logic                     := 'X';             -- clk
    rx_156_25_clk                   : in  std_logic                     := 'X';             -- clk
    csr_clk                         : in  std_logic                     := 'X';             -- clk
    csr_rst_n                       : in  std_logic                     := 'X';             -- reset_n
    tx_rst_n                        : in  std_logic                     := 'X';             -- reset_n
    rx_rst_n                        : in  std_logic                     := 'X';             -- reset_n
    avalon_st_tx_startofpacket      : in  std_logic                     := 'X';             -- startofpacket
    avalon_st_tx_endofpacket        : in  std_logic                     := 'X';             -- endofpacket
    avalon_st_tx_valid              : in  std_logic                     := 'X';             -- valid
    avalon_st_tx_data               : in  std_logic_vector(63 downto 0) := (others => 'X'); -- data
    avalon_st_tx_empty              : in  std_logic_vector(2 downto 0)  := (others => 'X'); -- empty
    avalon_st_tx_error              : in  std_logic                     := 'X';             -- error
    avalon_st_tx_ready              : out std_logic;                                        -- ready
    avalon_st_pause_data            : in  std_logic_vector(1 downto 0)  := (others => 'X'); -- data
    xgmii_tx                        : out std_logic_vector(71 downto 0);                    -- data
    avalon_st_txstatus_valid        : out std_logic;                                        -- valid
    avalon_st_txstatus_data         : out std_logic_vector(39 downto 0);                    -- data
    avalon_st_txstatus_error        : out std_logic_vector(6 downto 0);                     -- error
    xgmii_rx                        : in  std_logic_vector(71 downto 0) := (others => 'X'); -- data
    link_fault_status_xgmii_rx_data : out std_logic_vector(1 downto 0);                     -- data
    avalon_st_rx_data               : out std_logic_vector(63 downto 0);                    -- data
    avalon_st_rx_startofpacket      : out std_logic;                                        -- startofpacket
    avalon_st_rx_valid              : out std_logic;                                        -- valid
    avalon_st_rx_empty              : out std_logic_vector(2 downto 0);                     -- empty
    avalon_st_rx_error              : out std_logic_vector(5 downto 0);                     -- error
    avalon_st_rx_ready              : in  std_logic                     := 'X';             -- ready
    avalon_st_rx_endofpacket        : out std_logic;                                        -- endofpacket
    avalon_st_rxstatus_valid        : out std_logic;                                        -- valid
    avalon_st_rxstatus_data         : out std_logic_vector(39 downto 0);                    -- data
    avalon_st_rxstatus_error        : out std_logic_vector(6 downto 0)                      -- error
  );
  end component ip_arria10_mac_10g;

  component tech_transceiver_arria10_48 is
  generic (
    g_nof_channels      : natural := 48
  );
  port (
    clk                     : in  std_logic;
    reset_p                 : in  std_logic;
    refclk                  : in  std_logic;
    tx_clk_arr              : out std_logic_vector(g_nof_macs-1 DOWNTO 0);
    rx_clk_arr              : out std_logic_vector(g_nof_macs-1 DOWNTO 0);
    tx_serial_data          : out std_logic_vector(g_nof_macs-1 downto 0);
    rx_serial_data          : in  std_logic_vector(g_nof_macs-1 downto 0);
    tx_parallel_data        : in  std_logic_vector(64*g_nof_macs-1 downto 0);
    rx_parallel_data        : out std_logic_vector(64*g_nof_macs-1 downto 0);
    tx_control              : in  std_logic_vector(8*g_nof_macs-1 downto 0);
    rx_control              : out std_logic_vector(8*g_nof_macs-1 downto 0)
  );
  end component;

BEGIN

  ---------------------------------------------------------------------------------------
  -- Clocks and reset
  ---------------------------------------------------------------------------------------
  gen_tx_clk_stratixiv : if g_technology=c_tech_stratixiv generate	
    gen_tx_clk_arr: FOR i IN g_nof_macs-1 DOWNTO 0 GENERATE
      tx_clk_arr(i) <= tr_clk;
    END GENERATE;
  end generate;

  tx_rst_n_arr <= NOT tx_rst_arr;
  rx_rst_n_arr <= NOT rx_rst_arr;

  mm_rst_n <= NOT mm_rst;

  ---------------------------------------------------------------------------------------
  -- TX FIFO: dp_clk -> tx_clk
  ---------------------------------------------------------------------------------------
  gen_dp_fifo_dc_tx : FOR i IN g_nof_macs-1 DOWNTO 0 GENERATE
    u_dp_fifo_dc_tx : ENTITY dp_lib.dp_fifo_dc
    GENERIC MAP (
      g_data_w    => c_xgmii_data_w
    )
    PORT MAP (
      wr_rst      => dp_rst,
      wr_clk      => dp_clk,
      rd_rst      => tx_rst_arr(i),
      rd_clk      => tx_clk_arr(i),
  
      snk_out     => snk_out_arr(i),
      snk_in      => snk_in_arr(i),
  
      src_in      => dp_fifo_dc_tx_src_in_arr(i), 
      src_out     => dp_fifo_dc_tx_src_out_arr(i)
    );
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- TX packetizing: Add a header to the stream
  ---------------------------------------------------------------------------------------
  gen_dp_hdr_insert : FOR i IN 0 TO g_nof_macs-1 GENERATE
    u_dp_hdr_insert : ENTITY dp_lib.dp_hdr_insert
    GENERIC MAP (
      g_data_w          => c_xgmii_data_w,
      g_symbol_w        => c_byte_w,
      g_hdr_nof_words   => c_nof_header_words,
      g_internal_bypass => NOT g_use_hdr_ram,
      g_init_hdr        => g_hdr_ram_init_file,
      g_dp_on_at_init   => g_hdr_release_at_init
    )
    PORT MAP (
      mm_rst      => mm_rst,
      mm_clk      => mm_clk, 
                             
      st_rst      => tx_rst_arr(i), 
      st_clk      => tx_clk_arr(i), 
  
      reg_mosi    => reg_hdr_insert_mosi_arr(i),  
      ram_mosi    => ram_hdr_insert_mosi_arr(i),
                             
      snk_out     => dp_fifo_dc_tx_src_in_arr(i),
      snk_in      => dp_fifo_dc_tx_src_out_arr(i),
                             
      src_in      => dp_hdr_insert_src_in_arr(i),
      src_out     => dp_hdr_insert_src_out_arr(i)
    );
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- TX: Remove the word-alignment padding that precedes the DP frame
  ---------------------------------------------------------------------------------------
  gen_dp_pad_remove : FOR i IN 0 TO g_nof_macs-1 GENERATE
    u_dp_pad_remove : ENTITY dp_lib.dp_pad_remove
    GENERIC MAP (
      g_data_w          => c_xgmii_data_w,
      g_symbol_w        => c_byte_w,
      g_nof_padding     => c_word_alignment_padding_nof_bytes,
      g_internal_bypass => NOT g_word_alignment_padding
    )
    PORT MAP (
      rst         => tx_rst_arr(i),
      clk         => tx_clk_arr(i),

      snk_out     => dp_hdr_insert_src_in_arr(i),
      snk_in      => dp_hdr_insert_src_out_arr(i),

      src_in      => dp_pad_remove_src_in_arr(i), --,
      src_out     => dp_pad_remove_src_out_arr(i) --
    );
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- TX: Add one empty word at the tail; the MAC will insert its CRC there.
  --     This is only required if there are not 4 or more empty symbols in the last word.
  --     If a 6-byte padding was removed from the SOP word, weĺl end up with 6 empty 
  --     bytes in the EOP word which is already sufficient for CRC insertion.
  ---------------------------------------------------------------------------------------
  gen_empty_at_eop : IF g_word_alignment_padding = FALSE GENERATE

    gen_dp_concat_snk : FOR i IN 0 TO g_nof_macs-1 GENERATE
      dp_concat_snk_in_2arr(i)(1) <= dp_pad_remove_src_out_arr(i);
      dp_pad_remove_src_in_arr(i) <= dp_concat_snk_out_2arr(i)(1);
    END GENERATE;

    gen_dp_ready : FOR i IN 0 TO g_nof_macs-1 GENERATE -- Add flow control to dp_concat_snk_in
      u_dp_ready : ENTITY dp_lib.dp_ready
      PORT MAP (
        rst         => tx_rst_arr(i),
        clk         => tx_clk_arr(i),
  
        snk_in      => c_empty_crc_sosi,
  
        src_in      => dp_concat_snk_out_2arr(i)(0),
        src_out     => dp_concat_snk_in_2arr(i)(0)
      );
    END GENERATE;
  
    gen_dp_concat : FOR i IN 0 TO g_nof_macs-1 GENERATE
      u_dp_concat : ENTITY dp_lib.dp_concat
      GENERIC MAP (
        g_data_w    => c_xgmii_data_w,
        g_symbol_w  => c_byte_w
      )
      PORT MAP (
        rst         => tx_rst_arr(i),
        clk         => tx_clk_arr(i),
        snk_out_arr => dp_concat_snk_out_2arr(i),
        snk_in_arr  => dp_concat_snk_in_2arr(i),
        src_in      => dp_concat_src_in_arr(i),
        src_out     => dp_concat_src_out_arr(i)
      );
 
      dp_concat_src_in_arr(i) <= dp_fifo_fill_snk_out_arr(i);
 
      -- Set the empty field at the eop, dp_concat does not preserve it
      proc_empty : PROCESS(dp_concat_src_out_arr(i))
      BEGIN
        dp_fifo_fill_snk_in_arr(i) <= dp_concat_src_out_arr(i);
        dp_fifo_fill_snk_in_arr(i).empty <= (OTHERS=>'0');
  
        IF dp_concat_src_out_arr(i).eop = '1' THEN
          dp_fifo_fill_snk_in_arr(i).empty <= (OTHERS=>'1');       
        END IF;
      END PROCESS;
    END GENERATE;

  END GENERATE;

  gen_another_name : IF g_word_alignment_padding = TRUE GENERATE

    gen_dp_concat_snk : FOR i IN 0 TO g_nof_macs-1 GENERATE
      dp_fifo_fill_snk_in_arr(i) <= dp_pad_remove_src_out_arr(i);
      dp_pad_remove_src_in_arr(i) <=  dp_fifo_fill_snk_out_arr(i);
    END GENERATE;

  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- TX: FIFO so we can deliver packets to the MAC fast enough
  ---------------------------------------------------------------------------------------
  gen_dp_fifo_fill : FOR i IN 0 TO g_nof_macs-1 GENERATE
    u_dp_fifo_fill : ENTITY dp_lib.dp_fifo_fill
    GENERIC MAP (
      g_data_w      => c_xgmii_data_w,
      g_bsn_w       => 0,
      g_empty_w     => 3,
      g_channel_w   => 0,
      g_error_w     => 0,
      g_use_bsn     => FALSE,
      g_use_empty   => TRUE,
      g_use_channel => FALSE,
      g_use_error   => FALSE,
      g_use_sync    => FALSE,
      g_fifo_fill   => g_pkt_len, --Release packet only when available
      g_fifo_size   => g_pkt_len+c_fifo_margin
    )
    PORT MAP (
      rst         => tx_rst_arr(i),
      clk         => tx_clk_arr(i),

      snk_out     => dp_fifo_fill_snk_out_arr(i), --dp_concat_src_in_arr(i),
      snk_in      => dp_fifo_fill_snk_in_arr(i),

      src_in      => dp_fifo_fill_src_in_arr(i),
      src_out     => dp_fifo_fill_src_out_arr(i)
    );
  END GENERATE;

  dp_fifo_fill_src_in_arr <= dp_pad_insert_snk_out_arr WHEN g_lpbk_sosi=TRUE ELSE dp_latency_adapter_snk_out_arr;

  ---------------------------------------------------------------------------------------
  -- 10GbE MAC IP
  ---------------------------------------------------------------------------------------
  dp_latency_adapter_snk_in_arr <= dp_fifo_fill_src_out_arr WHEN g_lpbk_sosi=FALSE ELSE (OTHERS=>c_dp_sosi_rst);

  gen_dp_latency_adapter : FOR i IN 0 TO sel_a_b(g_lpbk_sosi, 0, g_nof_macs)-1 GENERATE
    u_dp_latency_adapter : ENTITY dp_lib.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 1,
      g_out_latency => 0
    )
    PORT MAP (
      rst       => tx_rst_arr(i),
      clk       => tx_clk_arr(i),
      -- ST sink
      snk_out   => dp_latency_adapter_snk_out_arr(i),
      snk_in    => dp_latency_adapter_snk_in_arr(i),
      -- ST source
      src_in    => mac_10g_snk_out_arr(i),
      src_out   => mac_10g_snk_in_arr(i)
    );
  END GENERATE;

  ---------------------- Stratix iv -------------------------

  gen_ip_stratixiv : if g_technology=c_tech_stratixiv generate
    gen_mac_10g : FOR i IN 0 TO sel_a_b(g_lpbk_sosi, 0, g_nof_macs)-1 GENERATE
    
      u_tech_mac_10g : ENTITY tech_mac_10g_lib.tech_mac_10g
      GENERIC MAP (
        g_technology          => g_technology,
        g_pre_header_padding  => FALSE
      )
      PORT MAP (
        -- MM
        mm_clk           => mm_clk,
        mm_rst           => mm_rst,
        csr_mosi         => reg_mac_mosi_arr(i),     -- CSR = control status register
        csr_miso         => reg_mac_miso_arr(i),
      
        -- ST
        tx_clk           => tx_clk_arr(i),           -- 156.25 MHz local reference
        tx_rst           => tx_rst_arr(i),
        tx_snk_in        => mac_10g_snk_in_arr(i),   -- 64 bit data
        tx_snk_out       => mac_10g_snk_out_arr(i),
        
        rx_clk           => rx_clk_arr(i),           -- 156.25 MHz from rx phy
        rx_rst           => rx_rst_arr(i),
        rx_src_out       => mac_10g_src_out_arr(i),  -- 64 bit data
        rx_src_in        => mac_10g_src_in_arr(i), 
        
        -- XGMII
        xgmii_tx_data    => mac_xgmii_tx_dc_arr(i),  -- 72 bit
        xgmii_rx_data    => mac_xgmii_rx_dc_arr(i)   -- 72 bit
      );
      
    END GENERATE;

    mac_10g_src_in_arr <= dp_pad_insert_snk_out_arr WHEN g_use_hdr_ram=TRUE ELSE dp_fifo_dc_rx_snk_out_arr;

    ---------------------------------------------------------------------------------------
    -- XGMII loopback: tr_xaui <-> ][ <-> tr_10GbE
    ---------------------------------------------------------------------------------------
    gen_lpbk_xgmii: IF g_lpbk_xgmii = TRUE GENERATE
      -- 10GbE side loopback: [ <-> tr_10GbE
      mac_xgmii_rx_dc_arr  <= mac_xgmii_tx_dc_arr;
      -- XAUI side loopback: tr_xaui <-> ]
      xaui_xgmii_rx_dc_arr <= xaui_xgmii_tx_dc_arr;
    END GENERATE;

    no_lpbk_xgmii: IF g_lpbk_xgmii = FALSE GENERATE
      xaui_xgmii_tx_dc_arr <= mac_xgmii_tx_dc_arr;
      mac_xgmii_rx_dc_arr  <= xaui_xgmii_rx_dc_arr;
    END GENERATE;

    u_tr_xaui : ENTITY tr_xaui_lib.tr_xaui
    GENERIC MAP (
      g_technology            => g_technology,
      g_sim                   => g_sim,
      g_sim_level             => g_sim_level,
      g_use_xgmii             => TRUE, -- Use XGMII direct
      g_nof_xaui              => g_nof_macs,
      g_mdio                  => g_use_mdio,
      g_mdio_epcs_dis         => g_mdio_epcs_dis
    )
    PORT MAP (
      -- Transceiver PLL reference clock   
      tr_clk                  => tr_clk,
  
      -- Calibration & reconfig clock
      cal_rec_clk             => cal_rec_clk,
   
      -- MM clock for register of optional MDIO master
      mm_clk                  => mm_clk,
      mm_rst                  => mm_rst,
  
      -- Streaming RX interfaces
      rx_clk                  => rx_clk_arr,
      rx_rst                  => rx_rst_arr,
  
      -- Streaming TX interfaces
      tx_clk                  => tx_clk_arr,
      tx_rst                  => tx_rst_arr,
  
      -- Direct XGMII interface
      xgmii_tx_dc_arr         => xaui_xgmii_tx_dc_arr,
      xgmii_rx_dc_arr         => xaui_xgmii_rx_dc_arr,
       
      --Serial I/O
      xaui_rx                 => xaui_rx_arr,
      xaui_tx                 => xaui_tx_arr,   
  
      -- XAUI PHY IP MM control/status
      xaui_mosi               => xaui_mosi,
      xaui_miso               => xaui_miso,
        
      -- MDIO master = mm slave
      mdio_mosi_arr           => mdio_mosi_arr(g_nof_macs-1 DOWNTO 0),
      mdio_miso_arr           => mdio_miso_arr(g_nof_macs-1 DOWNTO 0),
  
      -- MDIO External clock and serial data.
      mdio_rst                => mdio_rst,
      mdio_mdc                => mdio_mdc_arr,
      mdio_mdat_in            => mdio_mdat_in_arr,
      mdio_mdat_oen           => mdio_mdat_oen_arr
    );
  END GENERATE;


  ---------------------- Arria 10 -------------------------

  gen_ip_arria10 : if g_technology=c_tech_arria10 generate

	  gen_mac_10g : FOR i IN 0 TO sel_a_b(g_lpbk_sosi, 0, g_nof_macs)-1 generate

      mac_10g_snk_out_arr(i).xon <= tx_rst_n_arr(i);

      u0 : ip_arria10_mac_10g
      port map (
        csr_read                    => reg_mac_mosi_arr(i).rd,
        csr_write                   => reg_mac_mosi_arr(i).wr,
        csr_writedata               => reg_mac_mosi_arr(i).wrdata(c_word_w-1 downto 0),
        csr_readdata                => reg_mac_miso_arr(i).rddata(c_word_w-1 downto 0),
        csr_waitrequest             => reg_mac_miso_arr(i).waitrequest,
        csr_address                 => reg_mac_mosi_arr(i).address(c_mac_mm_addr_w-1 downto 0),
        tx_312_5_clk                => tx_clk_arr(i),
        tx_156_25_clk               => tx_clk_arr(i),
        rx_312_5_clk                => rx_clk_arr(i),
        rx_156_25_clk               => rx_clk_arr(i),
        csr_clk                     => mm_clk,
        csr_rst_n                   => mm_rst_n,
        tx_rst_n                    => tx_rst_n_arr(i),
        rx_rst_n                    => rx_rst_n_arr(i),
        avalon_st_tx_startofpacket  => mac_10g_snk_in_arr(i).sop,
        avalon_st_tx_endofpacket    => mac_10g_snk_in_arr(i).eop,
        avalon_st_tx_valid          => mac_10g_snk_in_arr(i).valid,
        avalon_st_tx_data           => mac_10g_snk_in_arr(i).data(c_xgmii_data_w-1 downto 0),
        avalon_st_tx_empty          => mac_10g_snk_in_arr(i).empty(2 downto 0),
        avalon_st_tx_error          => mac_10g_snk_in_arr(i).err(0),
        avalon_st_tx_ready          => mac_10g_snk_out_arr(i).ready,
        avalon_st_pause_data        => (others => '0'),
        xgmii_tx                    => mac_xgmii_tx_dc_arr(i),
        avalon_st_txstatus_valid    => open,
        avalon_st_txstatus_data     => open,
        avalon_st_txstatus_error    => open,
        xgmii_rx                    => mac_xgmii_rx_dc_arr(i),
        link_fault_status_xgmii_rx_data => open,
        avalon_st_rx_data           => mac_10g_src_out_arr(i).data(c_xgmii_data_w-1 downto 0),
        avalon_st_rx_startofpacket  => mac_10g_src_out_arr(i).sop,
        avalon_st_rx_valid          => mac_10g_src_out_arr(i).valid,
        avalon_st_rx_empty          => mac_10g_src_out_arr(i).empty(2 downto 0),
        avalon_st_rx_error          => mac_10g_src_out_arr(i).err(5 downto 0),
        avalon_st_rx_ready          => mac_10g_src_in_arr(i).ready,
        avalon_st_rx_endofpacket    => mac_10g_src_out_arr(i).eop,
        avalon_st_rxstatus_valid    => open,
        avalon_st_rxstatus_data     => open,
        avalon_st_rxstatus_error    => open
      );
    end generate;


    gen_lpbk_xgmii: IF g_lpbk_xgmii = TRUE generate
      -- 10GbE side loopback: [ <-> tr_10GbE
      mac_xgmii_rx_dc_arr  <= mac_xgmii_tx_dc_arr;
      -- XAUI side loopback: tr_xaui <-> ]
      tx_parallel_data     <= rx_parallel_data;
      tx_control           <= rx_control; 
      tx_data_valid        <= rx_data_valid;
      xaui_xgmii_rx_dc_arr <= xaui_xgmii_tx_dc_arr;
    end generate;

    no_lpbk_xgmii: IF g_lpbk_xgmii = FALSE generate
      gen_arrays : for i in 0 to g_nof_macs-1 generate	    
        tx_parallel_data(64*i+63 downto 64*i)  <= xgmii_d(mac_xgmii_tx_dc_arr(i));
        tx_control(8*i+7 downto 8*i)           <= xgmii_c(mac_xgmii_tx_dc_arr(i));
        mac_xgmii_rx_dc_arr(i)                 <= xgmii_dc(rx_parallel_data(64*i+63 downto 64*i),
                                                  rx_control(8*i+7 downto 8*i));
      end generate gen_arrays;
    end generate;


    u_transceiver: tech_transceiver_arria10_48 
    generic map (
      g_nof_channels     => g_nof_macs
    )
    port map(
      clk		 => mm_clk,
      reset_p            => mm_rst,
      refclk             => tr_clk,
      tx_clk_arr         => tx_clk_arr,
      rx_clk_arr         => rx_clk_arr,
      tx_serial_data     => tx_serial_data_internal,
      rx_serial_data     => rx_serial_data_internal,
      tx_parallel_data   => tx_parallel_data,
      rx_parallel_data   => rx_parallel_data,
      tx_control         => tx_control,
      rx_control         => rx_control
    );
  end generate;




  ---------------------------------------------------------------------------------------
  -- XAUI loopback: i/o pins <-> ][ <-> tr_xaui
  ---------------------------------------------------------------------------------------
  gen_lpbk_xaui: IF g_lpbk_xaui = TRUE GENERATE
    -- XAUI side loopback: [ <-> tr_xaui
    xaui_rx_arr <= xaui_tx_arr;
    -- I/O side loopback: i/o pins <-> ]
    xaui_tx_out_arr <= xaui_rx_in_arr;
    -- Same for arria10
    tx_serial_data <= rx_serial_data;
    rx_serial_data_internal <= tx_serial_data_internal;
  END GENERATE;

  no_lpbk_xaui: IF g_lpbk_xaui = FALSE GENERATE
    xaui_tx_out_arr <= xaui_tx_arr;
    xaui_rx_arr     <= xaui_rx_in_arr;
    tx_serial_data  <= tx_serial_data_internal;
    rx_serial_data_internal  <= rx_serial_data;
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- RX: Insert padding to align to 64b boundary
  ---------------------------------------------------------------------------------------
  dp_pad_insert_snk_in_arr <= dp_fifo_fill_src_out_arr WHEN g_lpbk_sosi=TRUE ELSE mac_10g_src_out_arr;

  gen_dp_pad_insert : FOR i IN 0 TO g_nof_macs-1 GENERATE
    u_dp_pad_insert : ENTITY dp_lib.dp_pad_insert
    GENERIC MAP (
      g_data_w          => c_xgmii_data_w,
      g_symbol_w        => c_byte_w,
      g_nof_padding     => c_word_alignment_padding_nof_bytes,
      g_internal_bypass => NOT g_word_alignment_padding
    )
    PORT MAP (
      rst         => rx_rst_arr(i),
      clk         => rx_clk_arr(i),

      snk_out     => dp_pad_insert_snk_out_arr(i),
      snk_in      => dp_pad_insert_snk_in_arr(i),

      src_in      => dp_frame_remove_snk_out_arr(i),
      src_out     => dp_frame_remove_snk_in_arr(i)
    );
  END GENERATE;
 
  ---------------------------------------------------------------------------------------
  -- RX: Unframe: remove header (and CRC if phy link is used) from DP packets
  ---------------------------------------------------------------------------------------        
  gen_dp_frame_remove : FOR i IN 0 TO g_nof_macs-1 GENERATE
    u_dp_frame_remove : ENTITY dp_lib.dp_frame_remove
    GENERIC MAP (            
      g_data_w          => c_xgmii_data_w,
      g_symbol_w        => c_byte_w,
      g_hdr_nof_words   => c_nof_header_words,
      g_tail_nof_words  => sel_a_b(c_remove_crc, c_nof_crc_words, 0),
      g_snk_latency     => 0,
      g_internal_bypass => NOT g_use_hdr_ram
    )
    PORT MAP (
      mm_rst      => mm_rst, 
      mm_clk      => mm_clk, 
                             
      st_rst      => rx_rst_arr(i), 
      st_clk      => rx_clk_arr(i), 
                             
      snk_out     => dp_frame_remove_snk_out_arr(i),
      snk_in      => dp_frame_remove_snk_in_arr(i),
  
      -- dp_frame_remove uses hdr_remove internally
      sla_in      => ram_hdr_remove_mosi_arr(i),
      sla_out     => ram_hdr_remove_miso_arr(i),
  
      src_in      => dp_fifo_dc_rx_snk_out_arr(i),
      src_out     => dp_fifo_dc_rx_snk_in_arr(i)
    );
  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- RX FIFO: rx_clk -> dp_clk
  ---------------------------------------------------------------------------------------
  gen_dp_fifo_dc_rx : FOR i IN g_nof_macs-1 DOWNTO 0 GENERATE
    u_dp_fifo_dc_rx : ENTITY dp_lib.dp_fifo_dc
    GENERIC MAP (
      g_data_w    => c_xgmii_data_w
    )
    PORT MAP (
      wr_rst      => rx_rst_arr(i),
      wr_clk      => rx_clk_arr(i),
      rd_rst      => dp_rst,
      rd_clk      => dp_clk,
  
      snk_out     => dp_fifo_dc_rx_snk_out_arr(i),
      snk_in      => dp_fifo_dc_rx_snk_in_arr(i),
  
      src_in      => dp_fifo_dc_rx_src_in_arr(i), 
      src_out     => dp_fifo_dc_rx_src_out_arr(i)
    );   
  END GENERATE;

  -----------------------------------------------------------------------------
  -- RX XON frame control
  -----------------------------------------------------------------------------
  gen_dp_xonoff : FOR i IN g_nof_macs-1 DOWNTO 0 GENERATE
    u_dp_xonoff : ENTITY dp_lib.dp_xonoff
      PORT MAP (
        rst      => dp_rst,
        clk      => dp_clk,
             
        in_siso  => dp_fifo_dc_rx_src_in_arr(i),      
        in_sosi  => dp_fifo_dc_rx_src_out_arr(i),
        
        out_siso => src_in_arr(i),
        out_sosi => src_out_arr(i)    
      );   
  END GENERATE;

  -----------------------------------------------------------------------------
  -- MM bus muxes
  -----------------------------------------------------------------------------
  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_macs,
    g_mult_addr_w => c_mac_mm_addr_w
  )
  PORT MAP (
    mosi     => reg_mac_mosi,
    miso     => reg_mac_miso,
    mosi_arr => reg_mac_mosi_arr,
    miso_arr => reg_mac_miso_arr
  );

  u_common_mem_mux_hdr_ins_reg : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_macs,
    g_mult_addr_w => c_hdr_insert_reg_addr_w
  )
  PORT MAP (
    mosi     => reg_hdr_insert_mosi,
    mosi_arr => reg_hdr_insert_mosi_arr
  );

  u_common_mem_mux_hdr_ins_ram : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_macs,
    g_mult_addr_w => c_hdr_insert_ram_addr_w
  )
  PORT MAP (
    mosi     => ram_hdr_insert_mosi,
    mosi_arr => ram_hdr_insert_mosi_arr
  );

  u_common_mem_mux_hdr_rem_ram : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_macs,
    g_mult_addr_w => c_hdr_remove_ram_addr_w
  )
  PORT MAP (
    mosi     => ram_hdr_remove_mosi,
    miso     => ram_hdr_remove_miso,
    mosi_arr => ram_hdr_remove_mosi_arr,
    miso_arr => ram_hdr_remove_miso_arr
  );
 
END str;
