--------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, tr_nonbonded_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;

ENTITY sim_xaui IS
  GENERIC(
    g_nof_xaui        : NATURAL := 1
  );                      
  PORT (   
    -- Transceiver PLL reference clock   
    tr_clk                  : IN  STD_LOGIC;
    trc_rst                 : IN  STD_LOGIC;

    -- Calibration & reconfig clock
    cal_rec_clk             : IN  STD_LOGIC;
 
    tx_clk                  : IN  STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
    rx_clk                  : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);
   
    crc_rx_ready            : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0); --crc = synchronous to Cal_Rec_Clk
    crc_tx_ready            : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

    a_rx_channelaligned     : OUT STD_LOGIC_VECTOR(g_nof_xaui-1 DOWNTO 0);

    xgmii_tx_dc             : IN  t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);
    xgmii_rx_dc             : OUT t_xgmii_dc_arr(g_nof_xaui-1 DOWNTO 0);

    xaui_rx                 : IN  t_xaui_arr(g_nof_xaui-1 DOWNTO 0);
    xaui_tx                 : OUT t_xaui_arr(g_nof_xaui-1 DOWNTO 0)
  );
END sim_xaui;


ARCHITECTURE wrap OF sim_xaui IS  

  CONSTANT c_xaui_serdes_data_w    : NATURAL := 16;
  CONSTANT c_xaui_serdes_ctrl_w    : NATURAL := c_xaui_serdes_data_w/c_byte_w;
  CONSTANT c_xaui_serdes_line_rate : NATURAL := 3125;

  --XGMII control bits (one for each XGMII lane):
  SIGNAL xgmii_tx_c                : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_c                : t_xgmii_c_arr(g_nof_xaui-1 DOWNTO 0);
  
  --XGMII data
  SIGNAL xgmii_tx_d                : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);
  SIGNAL xgmii_rx_d                : t_xgmii_d_arr(g_nof_xaui-1 DOWNTO 0);
 
BEGIN

  rx_clk <= tx_clk;
 
  gen_nof_xaui : FOR i IN g_nof_xaui-1 DOWNTO 0 GENERATE

    xgmii_tx_d(i) <= xgmii_d(xgmii_tx_dc(i));
    xgmii_tx_c(i) <= xgmii_c(xgmii_tx_dc(i));

    xgmii_rx_dc(i) <= xgmii_dc(xgmii_rx_d(i), xgmii_rx_c(i));
    
    u_areset_tx_rdy : ENTITY common_lib.common_areset
    GENERIC MAP(
      g_rst_level => '0',
      g_delay_len => 10
    )
    PORT MAP(
      clk     => cal_rec_clk,
      in_rst  => '0',
      out_rst => crc_tx_ready(i)
    );    

    u_areset_rx_rdy : ENTITY common_lib.common_areset
    GENERIC MAP(
      g_rst_level => '0',
      g_delay_len => 20
    )
    PORT MAP(
      clk     => cal_rec_clk,
      in_rst  => '0',
      out_rst => crc_rx_ready(i)
    );   

    u_areset_rx_channelaligned : ENTITY common_lib.common_areset
    GENERIC MAP(
      g_rst_level => '0',
      g_delay_len => 30
    )
    PORT MAP(
      clk     => cal_rec_clk,
      in_rst  => '0',
      out_rst => a_rx_channelaligned(i)
    );    

    gen_serdes: FOR j IN c_nof_xaui_lanes-1 DOWNTO 0 GENERATE

      u_ser: ENTITY tr_nonbonded_lib.serializer
      GENERIC MAP (
        g_data_w    => c_xaui_serdes_data_w,
        g_line_rate => c_xaui_serdes_line_rate
      )
      PORT MAP (
          
        tr_clk             => tr_clk,  
        tr_rst             => trc_rst,
         
        tx_in_data         => xgmii_tx_d(i)(j*c_xaui_serdes_data_w+c_xaui_serdes_data_w-1 DOWNTO j*c_xaui_serdes_data_w),
        tx_in_ctrl         => xgmii_tx_c(i)(j*c_xaui_serdes_ctrl_w+c_xaui_serdes_ctrl_w-1 DOWNTO j*c_xaui_serdes_ctrl_w),
    
        tx_out             => xaui_tx(i)(j)
      );

      u_des: ENTITY tr_nonbonded_lib.deserializer 
      GENERIC MAP (
        g_data_w    => c_xaui_serdes_data_w,
        g_line_rate => c_xaui_serdes_line_rate
      )
      PORT MAP (
          
        tr_clk             => tr_clk,  
        tr_rst             => trc_rst,
         
        rx_out_data        => xgmii_rx_d(i)(j*c_xaui_serdes_data_w+c_xaui_serdes_data_w-1 DOWNTO j*c_xaui_serdes_data_w),
        rx_out_ctrl        => xgmii_rx_c(i)(j*c_xaui_serdes_ctrl_w+c_xaui_serdes_ctrl_w-1 DOWNTO j*c_xaui_serdes_ctrl_w),
    
        rx_in              => xaui_rx(i)(j)
      );

    END GENERATE;

  END GENERATE;
      
END wrap;

