--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;

ENTITY tech_transceiver_rx_align IS
  GENERIC(
    g_sim    : BOOLEAN;
    g_data_w : NATURAL;
    g_mbps   : NATURAL
  );
  PORT(      
    rx_clk             : IN  STD_LOGIC;
    rx_rst             : IN  STD_LOGIC;
    rx_align_en_in     : IN  STD_LOGIC;
    rx_align_en_out    : OUT STD_LOGIC;
    rxc_rx_aligned     : OUT STD_LOGIC;
    rx_state           : OUT STD_LOGIC_VECTOR(1 DOWNTO 0)
   );  
END tech_transceiver_rx_align;

ARCHITECTURE rtl OF tech_transceiver_rx_align IS

  CONSTANT c_rx_clk_freq : NATURAL := (g_mbps / g_data_w) / 10 * 8 *1000000;

  CONSTANT c_timeout_1ms : NATURAL := c_rx_clk_freq/1000;
  CONSTANT c_timeout_sim : NATURAL := 100;
  CONSTANT c_timeout     : NATURAL := sel_a_b(g_sim, c_timeout_sim, 200*c_timeout_1ms);
  CONSTANT c_timeout_w   : NATURAL := ceil_log2(c_timeout);

  TYPE t_state_enum IS (s_init, s_wait, s_word_align, s_deassert_word_align, s_byte_order, s_deassert_byte_order, s_rx_aligned);

  SIGNAL state                    : t_state_enum;
  SIGNAL nxt_state                : t_state_enum; 

  SIGNAL i_rx_align_en_out        : STD_LOGIC;
  SIGNAL nxt_rx_align_en_out      : STD_LOGIC;

  SIGNAL i_rxc_rx_aligned         : STD_LOGIC;
  SIGNAL nxt_rxc_rx_aligned       : STD_LOGIC;

  SIGNAL cycle_cnt                : STD_LOGIC_VECTOR(c_timeout_w-1 DOWNTO 0);
  SIGNAL nxt_cycle_cnt            : STD_LOGIC_VECTOR(c_timeout_w-1 DOWNTO 0);

  SIGNAL nxt_rx_state             : STD_LOGIC_VECTOR(1 DOWNTO 0);
  SIGNAL i_rx_state               : STD_LOGIC_VECTOR(1 DOWNTO 0);

BEGIN 

  rx_state <= i_rx_state;

  rx_align_en_out <= i_rx_align_en_out;
  rxc_rx_aligned  <= i_rxc_rx_aligned;
 
  p_clk : PROCESS(rx_rst, rx_clk)
  BEGIN
    IF rx_rst='1' THEN
      state                <= s_init;  
      i_rx_align_en_out    <= '0';
      i_rxc_rx_aligned     <= '0';
      cycle_cnt            <= (OTHERS=>'0');
      i_rx_state           <= (OTHERS=>'0');
    ELSIF rising_edge(rx_clk) THEN
      state                <= nxt_state;
      i_rx_align_en_out    <= nxt_rx_align_en_out;
      i_rxc_rx_aligned     <= nxt_rxc_rx_aligned;
      cycle_cnt            <= nxt_cycle_cnt;
      i_rx_state           <= nxt_rx_state;
    END IF;
  END PROCESS;

  p_state : PROCESS(state, i_rx_align_en_out, i_rxc_rx_aligned, cycle_cnt, i_rx_state, rx_align_en_in)
  BEGIN  
    nxt_state           <= state;  
    nxt_rx_align_en_out <= i_rx_align_en_out;
    nxt_rxc_rx_aligned  <= i_rxc_rx_aligned;
    nxt_cycle_cnt       <= cycle_cnt;
    nxt_rx_state        <= i_rx_state;
 
    CASE state IS
      WHEN s_init => 
        nxt_state <= s_wait;
  
      WHEN s_wait =>
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_timeout, c_timeout_w) THEN
           nxt_state     <= s_word_align;
           nxt_cycle_cnt <= (OTHERS=>'0');
        END IF;
  
      WHEN s_word_align =>  -- First assertion of rx_align_en to perform word alignment
        nxt_rx_state        <= "01";
        nxt_rx_align_en_out <= '1';
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_timeout, c_timeout_w) THEN
           nxt_state     <= s_deassert_word_align;
           nxt_cycle_cnt <= (OTHERS=>'0');
        END IF;
  
      WHEN s_deassert_word_align =>  -- Deasssert rx_align_en, re-assert after delay
        nxt_rx_align_en_out <= '0';
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_timeout, c_timeout_w) THEN
           nxt_state     <= s_byte_order;
           nxt_cycle_cnt <= (OTHERS=>'0');
        END IF;
  
      WHEN s_byte_order => -- Assert rx_align_en a second time to order the bytes
        nxt_rx_state        <= "10";
        nxt_rx_align_en_out <= '1';
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_timeout, c_timeout_w) THEN
           nxt_state     <= s_deassert_byte_order;
           nxt_cycle_cnt <= (OTHERS=>'0');
        END IF;
  
     WHEN s_deassert_byte_order =>  -- Deasssert rx_align_en
        nxt_rx_align_en_out <= '0';
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1); 
        IF cycle_cnt = TO_UVEC(c_timeout, c_timeout_w) THEN
           nxt_state  <= s_rx_aligned;
           nxt_cycle_cnt <= (OTHERS=>'0');
        END IF;
  
      WHEN s_rx_aligned => -- Asserting rx_aligned on next cycle. 
        nxt_rx_state        <= "11";
        nxt_rx_align_en_out <= '0';
        nxt_rxc_rx_aligned  <= '1';
        IF rx_align_en_in = '1' THEN
          nxt_state           <= s_wait;
          nxt_rxc_rx_aligned  <= '0';
        END IF;
  
      WHEN OTHERS =>
        nxt_state <= s_init;
    END CASE;
  END PROCESS;
  
END rtl;

